_OBJ_CLASSES_BY_ALIAS = {}
_ALIASES_BY_OBJ_CLASS = {}


def add_alias(alias: str, obj_class_or_name: str):
    """
    Register an alias for a class or class name.
    This will be reflected on the field "py/object"
    parameters:
        alias: The alias to set for a given object type
        obj_class_or_name: Object class or class name.
    """
    obj_class_name = obj_class_or_name if isinstance(obj_class_or_name, str) else obj_class_or_name.object_type_name()
    _OBJ_CLASSES_BY_ALIAS[alias] = obj_class_name
    _ALIASES_BY_OBJ_CLASS[obj_class_name] = alias


def get_obj_class_name(name):
    return _OBJ_CLASSES_BY_ALIAS.get(name)


def get_alias(obj_class_name):
    return _ALIASES_BY_OBJ_CLASS.get(obj_class_name)



from .expressions import (
    BaseExpr,
    ObjectTypeExpr,
    SetExpr,
    Join,
    Projection,
    And,
    Or,
    Nor,
    Not,
    Field,
    SortExpr,
    SetOrInsertExpr,
)


class QueryHelper:
    """
    Utility class which allows making Mongo queries
    in a typed way, allowing abstracting implementation from
    real model.

    Having a some classes like this:

        class Item(JsonObject):
            product_id = Property(json="product-id")
            quantity = Property()

        class Invoice(JsonObject):
            items = Property(type=ArrayOf(Item))

    it is possible to create expressions like this:

        qhi = QueryHelper(Invoice)
        expression = qhi(qhi.items.product_id == "1000", qhi.items.quantity < 10)

    which will translate into this:

        {'items.product-id':'1000', 'items.quantity': {'$lt':10}}

    Another example:
        expression = (qhi.items.quantity == 100)
    
    while translate into this:

        {'items.quantity':100}

    All expressions are comparable with dictionaries or lists depending the case

    """

    def __init__(self, obj_type):
        self._obj_type = obj_type

    def name(self):
        return "py/state" if self._obj_type.__jsonpickle_format__ else ""

    def field_path(self):
        return [self] if self._obj_type.__jsonpickle_format__ else []

    def __getattr__(self, name):
        """
        Returns a field instance pointing to the expected field of the class
        """
        return Field(self, getattr(self._obj_type, name), name)

    def and_(self, *args):
        """
        $and operator
        """
        return And(self, args)

    def __call__(self, *args):
        """
        Join all expressions passed as parameter into one single dictionary,
        this is the default $and operation.
        """
        return Join(self, args)

    def project(self, *fields):
        """
        Builds a query projection expression dictionary.
        As input it receives a list of fields, for those which is required on the result, 
        or a tuple (field, False), for those fields not required on the result.
        """
        return Projection(fields)

    def sort(self, *sort_exprs):
        """
        Builds a list of (field, sort), by receiving a list of field.asc()/desc().
        """
        return list(map(SortExpr.expr, sort_exprs))

    def or_(self, *args):
        """
        $or operator
        """
        return Or(self, args)

    def nor(self, *args):
        """
        $nor operator
        """
        return Nor(self, args)

    def not_(self, arg):
        """
        $not operator
        """
        return Not(self, arg)

    def type_expr(self):
        """
        Return an expression for checking document's py/object value.
        This is translated into: {'py/object' : 'module.classname'}
        """
        return ObjectTypeExpr(self)

    def set(self, *args):
        return SetExpr(self, args)

    def set_or_insert(self, *args):
        return SetOrInsertExpr(self, args)



from podm.meta import CollectionOf
from abc import ABCMeta, abstractmethod
from collections.abc import Mapping


class BaseExpr(Mapping, metaclass=ABCMeta):
    """
    Interface for expressions.
    """

    @abstractmethod
    def expr(self):
        """
        Convert this expression into a dictionary representing
        a mongo query expression.
        """
        pass

    def keys(self):
        return self.expr().keys()

    def items(self):
        return self.expr().items()

    def values(self):
        return self.expr().values()

    def __getitem__(self, key):
        return self.expr()[key]

    def __len__(self):
        return 1

    def __hash__(self):
        return hash(self.expr())

    def __eq__(self, other):
        if isinstance(other, Expr):
            return self.expr() == other.expr
        elif isinstance(other, dict):
            return self.expr() == other
        return False

    def __str__(self):
        return str(self.expr())

    def __repr__(self):
        return str(self)

    def __contains__(self, key):
        return key in self.expr()

    def get(self, key, default=None):
        return self.expr().get(key, default)

    def __iter__(self):
        return iter(self.expr())


class Expr(BaseExpr):
    def __init__(self, field, value):
        self._field = field
        self._value = value


class Comparation(Expr):
    _comparator = None

    def expr(self):
        return {self._field.path(): {self._comparator: self._value}}


class Eq(Comparation):
    _comparator = "$eq"  # not used for this particular case in favor of {k:v}

    def expr(self):
        return {self._field.path(): self._value}


class Ne(Comparation):
    _comparator = "$ne"


class Lt(Comparation):
    _comparator = "$lt"


class Le(Comparation):
    _comparator = "$lte"


class Gt(Comparation):
    _comparator = "$gt"


class Ge(Comparation):
    _comparator = "$gte"


class In(Comparation):
    _comparator = "$in"


class Nin(Comparation):
    _comparator = "$nin"


class Exists(Comparation):
    _comparator = "$exists"


class Operation(BaseExpr):
    def __init__(self, parent, expressions):
        self._parent = parent
        self._expressions = expressions


class LogicalOperation(Operation):
    _operator = None

    def expr(self):
        return {self._operator: [i.expr() for i in self._expressions]}


class And(LogicalOperation):
    _operator = "$and"


class Join(And):
    def expr(self):
        result = {}

        for expr in self._expressions:
            result.update(expr.expr())

        return result


class Projection(BaseExpr):
    def __init__(self, fields):
        self._fields = fields

    def expr(self):
        return {f.path(): v for (f, v) in map(lambda x: (x, 1) if isinstance(x, BaseField) else x, self._fields)}


class Or(LogicalOperation):
    _operator = "$or"


class Nor(LogicalOperation):
    _operator = "$nor"


class Not(LogicalOperation):
    _operator = "$not"

    def expr(self):
        # Here it takes only first argument
        return {self._operator: self._expressions.expr()}


class SortExpr(BaseExpr):
    ASCENDING = 1
    DESCENDING = -1

    def __init__(self, field, val):
        self._field = field
        self._val = val

    def expr(self):
        return (self._field.path(), self._val)


class BaseSetExpr(BaseExpr):
    _operator = None

    def __init__(self, parent, set_dict):
        self._parent = parent
        self._set_dict = set_dict

    def expr(self):
        return {self._operator: dict([v.expr() for v in self._set_dict])}


class SetExpr(BaseSetExpr):
    _operator = "$set"


class SetOrInsertExpr(BaseSetExpr):
    _operator = "$setOrInsert"


class SetFieldExpr(BaseExpr):
    def __init__(self, field, val):
        self._field = field
        self._val = val

    def expr(self):
        return (self._field.path(), self._val)


class BaseField:
    def __init__(self, parent, field):
        self._parent = parent
        self._field = field

    def field_path(self):
        return self._parent.field_path() + [self]

    def _add_dot(self):
        return True

    def path(self):
        result = []
        for item in self.field_path():
            if result and item._add_dot():
                result.append(".")
            result.append(item.name())
        return "".join(result)

    def __getattr__(self, name):
        field_type = self._field.type

        if isinstance(field_type, CollectionOf):
            field_type = field_type.type

        return Field(self, getattr(field_type, name), name)

    def __getitem__(self, key):
        return ArrayItem(self, self._field, key)

    def __eq__(self, value):
        """
        $eq operator, translates expression a == b into {'a':b}
        """
        return Eq(self, value)

    def eq(self, value):
        """
        Same as __eq__
        """
        return Eq(self, value)

    def __ne__(self, value):
        """
        $ne operator, traslates expression a < b into {'a' : {'$lt': b} }
        """
        return Ne(self, value)

    def __lt__(self, value):
        """
        $lt operator, traslates expression a < b into {'a' : {'$lt': b} }
        """
        return Lt(self, value)

    def __le__(self, value):
        """
        $lte operator, traslates expression a <= b into {'a' : {'$lte': b} }
        """
        return Le(self, value)

    def __gt__(self, value):
        """
        $gt operator, traslates expression a > b into {'a' : {'$gt': b} }
        """
        return Gt(self, value)

    def __ge__(self, value):
        """
        $gte operator, traslates expression a > b into {'a' : {'$gte': b} }
        """
        return Ge(self, value)

    def in_(self, value):
        """
        $in operator
        """
        return In(self, value)

    def nin(self, value):
        """
        $nin operator
        """
        return Nin(self, value)

    def exists(self, val=True):
        """
        $exists operator, expects boolean argument.
        """
        return Exists(self, val)

    def asc(self):
        """
        ascending sort expression
        """
        return SortExpr(self, SortExpr.ASCENDING)

    def desc(self):
        """
        descending sort expression
        """
        return SortExpr(self, SortExpr.DESCENDING)

    def __call__(self, arg):
        return SetFieldExpr(self, arg)


class ArrayItem(BaseField):
    """
    Array item referencing.
    """

    def __init__(self, parent, field, key):
        super().__init__(parent, field)
        self._key = key

    def name(self):
        return "[" + (f"'{self._key}'" if isinstance(self._key, str) else str(self._key)) + "]"

    def _add_dot(self):
        return False


class ObjectTypeExpr(BaseField):
    def __init__(self, parent):
        super().__init__(parent, None)

    def name(self):
        return "py/object"

    def expr(self):
        return {self.name(): self._parent._obj_type.object_type_name()}


class Field(BaseField):
    """
    Field referencing.
    """

    def __init__(self, parent, field, name):
        super().__init__(parent, field)
        self._field_name = name

    def name(self):
        """
        Actual name of this field
        """
        return self._field.json or self._field_name



from .meta import CollectionOf, ArrayOf, MapOf
from enum import Enum


class SchemaBuilder:
    def __init__(self, obj_type):
        self._obj_type = obj_type

    def build(self, deep=True, base_schema_url=None):

        obj_properties = self._obj_type.properties()

        if deep:
            definitions = self._collect_definitions(obj_properties)
        else:
            definitions = {}

        properties = {p.json(): p.schema(definitions, deep, base_schema_url) for p in obj_properties.values()}
        required = [p.json() for p in obj_properties.values() if not p.allow_none()]

        schema = {"type": "object", "properties": {}}

        if self._obj_type.__add_type_identifier__:
            schema["properties"]["py/object"] = {"const": self._obj_type.object_type_name()}
        if self._obj_type.__jsonpickle_format__:
            schema["properties"]["py/state"] = {"$ref": "#/definitions/state"}
            definitions.update({"state": {"type": "object", "properties": properties}})
        else:
            schema["properties"].update(properties)

        if required:
            schema["required"] = required

        if definitions:
            schema["definitions"] = definitions

        return schema

    def _collect_definitions(self, properties):

        result = {}
        for prop in properties.values():
            field_type = prop.field_type()
            if isinstance(field_type, CollectionOf):
                inner_type = field_type.type
                result[inner_type.__name__] = inner_type.schema()
            elif (
                field_type
                and not self._primitive(field_type)
                and issubclass(field_type, object)
                and not issubclass(field_type, Enum)
            ):
                result[field_type.__name__] = field_type.schema()

        return result

    def _primitive(self, type):
        return type in [int, bool, float, str]



# vim:ts=4:sw=4:expandtab
import unittest
from podm import JsonObject, Property, Handler, Processor, ArrayOf, MapOf
from collections import OrderedDict
from datetime import datetime


class Entity(JsonObject):
    """
    A base class for the object model
    """

    oid = Property("oid", type=str)
    created = Property("created", default=datetime.now)


class Company(Entity):
    """
    This class represents a company.
    """

    company_name = Property("company-name", type=str)
    description = Property("description", type=str)

    def __init__(self, **kwargs):
        super(Company, self).__init__(**kwargs)
        # I use this field only for checking that the field "description"
        # is accessed from the getter instead of using default accessor.
        self._used_getter = False

    def get_description(self):
        # Set _used_getter to True if this getter has been called.
        self._used_getter = True
        return self._description


class Sector(Entity):
    employees = Property("employees", default=[])


class Employee(Entity):
    name = Property()


class DateTimeHandler(Handler):
    def encode(self, obj):
        return {
            "year": obj.year,
            "month": obj.month,
            "day": obj.day,
            "hour": obj.hour,
            "minute": obj.minute,
            "second": obj.second,
            "microsecond": obj.microsecond,
        }

    def decode(self, obj_data):
        return datetime(**obj_data)


class TestObject(JsonObject):

    date_time = Property("date-time", handler=DateTimeHandler(), default=datetime.now)

    def __init__(self, **kwargs):
        JsonObject.__init__(self, **kwargs)
        self._deserialized = False

    def _after_deserialize(self):
        self._deserialized = True


class TestObject2(JsonObject):
    property1 = Property()


class Child(JsonObject):
    property1 = Property()


class Parent(JsonObject):
    children = Property(type=ArrayOf(Child))



# vim:ts=4:sw=4:expandtab
import unittest
import json
from podm import JsonObject, Property, Handler, Processor, ArrayOf, MapOf
from collections import OrderedDict
from datetime import datetime
from .common import (
    Entity,
    Company,
    Sector,
    Employee,
    DateTimeHandler,
    TestObject,
    TestObject2,
    Child,
    Parent,
)
from enum import Enum


class Values(Enum):
    VALUE1 = 1
    VALUE2 = 2


class Type2(JsonObject):
    field_1 = Property(type=int, title="Field 1 title")
    field_2 = Property(type=Values, enum_as_str=True)
    field_3 = Property(type=str, pattern=".*")
    field_4 = Property(type=str, format="email")


class Type1(JsonObject):

    field_1 = Property(type=str)
    field_2 = Property(type=Type2)


class TestSchema(unittest.TestCase):
    def test_base_schema(self):
        schema = Company.schema()
        self.assertIn("type", schema)
        self.assertIn("properties", schema)
        self.assertIn("py/object", schema["properties"])
        self.assertIn("const", schema["properties"]["py/object"])
        self.assertEqual("test.common.Company", schema["properties"]["py/object"]["const"])
        self.assertIn("oid", schema["properties"])
        self.assertIn("type", schema["properties"]["oid"])
        self.assertEqual("string", schema["properties"]["oid"]["type"])
        self.assertIn("created", schema["properties"])
        self.assertEqual("object", schema["properties"]["created"]["type"])
        self.assertIn("company-name", schema["properties"])
        self.assertEqual("string", schema["properties"]["company-name"]["type"])
        self.assertIn("description", schema["properties"])
        self.assertEqual("string", schema["properties"]["description"]["type"])

    def test_compound(self):
        import json

        schema = Type1.schema()
        print(json.dumps(schema, indent=4))
        self.assertEqual(schema["type"], "object")
        self.assertEqual(schema["properties"]["field_2"], {"$ref": "#/definitions/Type2"})
        self.assertEqual(1, len(schema["definitions"]))
        self.assertIn("properties", schema["definitions"]["Type2"])
        self.assertIn("field_1", schema["definitions"]["Type2"]["properties"])
        self.assertIn("title", schema["definitions"]["Type2"]["properties"]["field_1"])
        self.assertIn("field_2", schema["definitions"]["Type2"]["properties"])
        self.assertIn("enum", schema["definitions"]["Type2"]["properties"]["field_2"])
        self.assertEqual(
            set(["VALUE1", "VALUE2"]), set(schema["definitions"]["Type2"]["properties"]["field_2"]["enum"]),
        )
        self.assertIn("field_3", schema["definitions"]["Type2"]["properties"])
        self.assertIn("pattern", schema["definitions"]["Type2"]["properties"]["field_3"])
        self.assertIn("field_4", schema["definitions"]["Type2"]["properties"])
        self.assertIn("format", schema["definitions"]["Type2"]["properties"]["field_4"])

    def test_schema_3(self):
        class Child(JsonObject):
            __add_type_identifier__ = False
            field1 = Property(type=str)

        class Parent(JsonObject):
            __add_type_identifier__ = False
            field1 = Property(type=str, allow_none=False, description="some field 1")
            field2 = Property(type=str, allow_none=False, description="some field 2")
            field3 = Property(
                type=str, allow_none=False, description="some field 3", schema_ref="http://schemas.org/fieldtype#"
            )

            keys = Property(default=dict, type=MapOf(Child))

        schema = Parent.schema()

        self.assertIn("patternProperties", schema["properties"]["keys"])
        self.assertIn(".*", schema["properties"]["keys"]["patternProperties"])
        self.assertIn("$ref", schema["properties"]["keys"]["patternProperties"][".*"])

        schema = Parent.schema(deep=False)
        self.assertIn("type", schema["properties"]["keys"]["patternProperties"][".*"])
        self.assertEqual("object", schema["properties"]["keys"]["patternProperties"][".*"]["type"])

        schema = Parent.schema(deep=False, base_schema_url="http://schemas.org/")
        self.assertIn("$ref", schema["properties"]["keys"]["patternProperties"][".*"])
        self.assertEqual("http://schemas.org//Child", schema["properties"]["keys"]["patternProperties"][".*"]["$ref"])

    def test_custom_schema(self):

        field_schema = {"type": "string", "format": "hostname"}

        class Item(JsonObject):
            field = Property(schema=field_schema)

        schema = Item.schema()

        self.assertEqual(field_schema, schema["properties"]["field"])



# vim:ts=4:sw=4:expandtab
import unittest
from podm import JsonObject, Property, Handler, Processor, ArrayOf, MapOf
from collections import OrderedDict
from datetime import datetime
from .common import (
    Entity,
    Company,
    Sector,
    Employee,
    DateTimeHandler,
    TestObject,
    TestObject2,
    Child,
    Parent,
)


class TestProcessor(unittest.TestCase):
    def test_processor(self):
        class UpperCaseProcessor(Processor):
            def when_from_dict(self, key, val):
                return key.lower(), val

            def when_to_dict(self, key, val):
                return key.upper(), val

        class TestObject(JsonObject):
            property1 = Property()

        test_object = TestObject()
        test_object.property1 = "hello!!"

        serialized = test_object.to_dict(processor=UpperCaseProcessor())
        self.assertIn("PROPERTY1", serialized)

        deserialized = TestObject.from_dict(serialized, processor=UpperCaseProcessor())
        self.assertEqual("hello!!", deserialized.property1)


if __name__ == "__main__":
    unittest.main()



# vim:ts=4:sw=4:expandtab
import unittest
from podm import JsonObject, ListStringHandler, Property


class Company:
    def __init__(self, string):
        self.string = string

    def __str__(self):
        return self.string


class Companies(JsonObject):
    company_list = Property("companies", default=[], handler=ListStringHandler(Company))


class TestHandlers(unittest.TestCase):
    def test_ListStringHandler_encoding(self):
        reference_list = ["company1", "company2", "company3"]
        data = {"companies": reference_list }
        companies = Companies.from_dict(data)

        self.assertTrue(all(isinstance(company, Company) for company in companies.company_list))
        self.assertCountEqual([str(c) for c in companies.company_list], reference_list)

    def test_ListStringHandler_decoding(self):
        reference_list = ["company1", "company2", "company3"]
        companies = Companies()
        companies.company_list = [Company(c) for c in reference_list]
        data = companies.to_dict()

        self.assertCountEqual(data["companies"], reference_list)




from podm import JsonObject, Property
from unittest import TestCase


class TestObject1(JsonObject):
    field1 = Property()

    def get_field2(self):
        return self._field2

    def set_field2(self, field2):
        self._field2 = field2

    field2 = property(get_field2, set_field2)

    def update_field2(self, val):
        self.field2 = val


class TestObject2(TestObject1):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._field2 = None


class TestBasic(TestCase):
    def test_mixed_properties(self):
        """
        Test for a bug when a regular property is in a base class,
        subclasses not being able to see it.
        """
        obj = TestObject2()
        obj.field2 = 1
        self.assertEqual(1, obj.field2)
        self.assertEqual(1, obj.get_field2())
        obj.update_field2(2)
        self.assertEqual(2, obj.field2)



# vim:ts=4:sw=4:expandtab
from .meta import Handler

class ListStringHandler(Handler):
    """
    Convert JSON list to and from list of given class_type

    class_type must implement str conversion to encode to JSON.
    """

    def __init__(self, class_type):
        self.class_type = class_type

    def encode(self, objects):
        return [str(obj) for obj in objects]

    def decode(self, string_list):
        return [self.class_type(string) for string in string_list]



from unittest import TestCase
from podm import JsonObject, Property


class TestObject(JsonObject):
    field1 = Property(default=1)
    field2 = Property(default=True)
    field3 = Property(default=False)
    field4 = Property(default=dict)
    field5 = Property(default=lambda: [1, 2, 3])


class TestDefaults(TestCase):
    def test_default_int(self):
        obj = TestObject()
        self.assertEqual(1, obj.field1)

    def test_default_bool_true(self):
        obj = TestObject()
        self.assertTrue(obj.field2)

    def test_default_bool_false(self):
        obj = TestObject()
        self.assertFalse(obj.field3)

    def test_default_dict(self):
        obj = TestObject()
        self.assertEqual({}, obj.field4)

    def test_default_list_factory(self):
        obj = TestObject()
        self.assertEqual([1, 2, 3], obj.field5)

    def test_constructors(self):
        obj = TestObject(field1=2, field2=False, field3=True, field4={"a": 1}, field5=[4, 5, 6])
        self.assertEqual(2, obj.field1)
        self.assertFalse(obj.field2)
        self.assertTrue(obj.field3)
        self.assertEqual({"a": 1}, obj.field4)
        self.assertEqual([4, 5, 6], obj.field5)



# vim:ts=4:sw=4:expandtab
__author__ = "Carlos Descalzi"

from .jsonobject import BaseJsonObject, JsonObject
from .meta import Handler, Property, ArrayOf, MapOf
from .processor import Processor
from .properties import PropertyHandler, RichPropertyHandler
from .validation import Validator, ValidationException
from .aliases import add_alias



from podm import JsonObject, Property, Validator, ValidationException
from unittest import TestCase
import traceback
import re


class ValidatorTest(JsonObject):
    __validate__ = True
    field_1 = Property(type=str, allow_none=False, validator="default")


class EmailValidator(Validator):
    def validate(self, obj, field, value):
        if not value:
            return f"{field} is mandatory"
        matched = re.match(r"[a-z0-9\.\-]+@[a-z0-9\.]+", value)
        if not matched:
            return f"Invalid email address {value}"
        return None


class ValidatorTest2(JsonObject):
    __validate__ = True
    email = Property(validator=EmailValidator())


class TestValidation(TestCase):
    def test_not_none(self):
        data = {"field_1": None}

        with self.assertRaises(ValidationException) as ctx:
            obj = ValidatorTest.from_dict(data)

        issues = ctx.exception.issues
        self.assertIsNotNone(issues)
        self.assertNotEqual(0, len(issues))

    def test_wrong_type(self):
        data = {"field_1": 1}

        with self.assertRaises(ValidationException) as ctx:
            obj = ValidatorTest.from_dict(data)

        issues = ctx.exception.issues
        self.assertIsNotNone(issues)
        self.assertNotEqual(0, len(issues))

    def test_right_type(self):
        data = {"field_1": "val"}
        try:
            obj = ValidatorTest.from_dict(data)
        except Exception as e:
            print(traceback.format_exc())
            self.assertTrue(False, str(e))

    def test_missing(self):
        data = {}
        with self.assertRaises(ValidationException) as ctx:
            obj = ValidatorTest.from_dict(data)

        issues = ctx.exception.issues
        self.assertIsNotNone(issues)
        self.assertNotEqual(0, len(issues))

    def test_custom_validator(self):

        data = {"email": None}
        with self.assertRaises(ValidationException) as ctx:
            obj = ValidatorTest2.from_dict(data)

        data = {"email": "xxxx"}
        with self.assertRaises(ValidationException) as ctx:
            obj = ValidatorTest2.from_dict(data)

        data = {"email": "xxxx@xxx.com"}
        try:
            obj = ValidatorTest2.from_dict(data)
        except Exception as e:
            self.assertTrue(False, str(e))



class AggregateHelper:
    """
    Not finished yet.
    """

    def match(self, *args):
        pass

    def sort(self, *args):
        pass

    def group(self, *args):
        pass

    def project(self, *args):
        pass

    def set(self, *args):
        pass

    def max(self, *args):
        pass

    def min(self, *args):
        pass

    def last(self, *args):
        pass

    def first(self, *args):
        pass

    def avg(self, *args):
        pass

    def divide(self, *args):
        pass

    def substract(self, *args):
        pass

    def if_null(self, *args):
        pass
